from ddnatlas.connector_collections import create_collection
from ddnatlas.columns import create_columns
from ddnatlas.data_connectors import create_data_connector
from ddnatlas.fields import create_fields
from ddnatlas.object_types import get_object_type_and_name, create_object_type
from ddnatlas.queries import create_queries
from ddnatlas.scalars import create_scalars
from ddnatlas.update_entities import update_entities
import logging


def generate_subgraph(supergraph, subgraph, connectors, data_connector_links,
                      object_types, models, include, exclude):
    logging.info('Starting to generate subgraph...')

    supergraph_name = supergraph['definition'].get('name', 'enterprise')
    subgraph_name = subgraph['definition']['name']
    subgraph_data_connectors = [item for item in connectors if item['definition']['subgraph'] == subgraph_name]
    entities = []
    data_connector_ids = []
    object_type_ids = []
    query_ids = []
    for subgraph_data_connector in subgraph_data_connectors:
        logging.info('Processing subgraph data connectors...')
        connector_name = subgraph_data_connector['definition']['name']
        link = [item for item in data_connector_links if item['definition']['name'] == connector_name][0]
        subgraph_data_connector['schema'] = link['definition']['schema']['schema']
        # CREATE SCALARS
        logging.info('Creating scalars...')
        scalar_ids = create_scalars(entities, supergraph_name, subgraph_name, connector_name, subgraph_data_connector)
        logging.info('Finished creating scalars. Total created: %d', len(scalar_ids))
        update_entities(entities, include, exclude)
        collection_ids = []
        for collection_name, object_schema in subgraph_data_connector['schema']['object_types'].items():
            # CREATE COLUMNS
            logging.info('Creating columns...')
            column_ids = create_columns(entities, supergraph_name, subgraph_name, connector_name, collection_name,
                                        object_schema)
            logging.info('Finished creating columns. Total created: %d', len(column_ids))
            update_entities(entities, include, exclude)
            # CREATE FIELDS
            logging.info('Creating fields...')
            object_type, object_name = get_object_type_and_name(object_types, connector_name, collection_name)
            field_ids = create_fields(supergraph_name, subgraph_name, object_name, object_type, entities)
            logging.info('Finished creating fields. Total created: %d', len(field_ids))
            update_entities(entities, include, exclude)
            # CREATE COLLECTION
            logging.info('Creating collections...')
            collection_ids.append(
                create_collection(subgraph_data_connector, collection_name, supergraph_name, subgraph_name, column_ids,
                                  entities))
            logging.info('Finished creating collections. Total created: %d', len(collection_ids))
            update_entities(entities, include, exclude)
            # CREATE OBJECT_TYPE
            logging.info('Creating object types...')
            object_qualified_name = create_object_type(supergraph_name, subgraph_name, object_name, field_ids, entities)
            object_type_ids.append(object_qualified_name)
            logging.info('Finished creating object types. Total created: %d', len(object_type_ids))
            update_entities(entities, include, exclude)
            # CREATE QUERIES
            logging.info('Creating queries...')
            data_connector_model = [item for item in models if
                                    item['definition']['source']['dataConnectorName'] == connector_name and
                                    item['definition']['source']['collection'] == collection_name][0]
            query_ids = query_ids + create_queries(supergraph_name, subgraph_name, object_qualified_name,
                                                   data_connector_model, entities)
            logging.info('Finished creating queries. Total created: %d', len(query_ids))
            update_entities(entities, include, exclude)
        # CREATE DATA CONNECTORS
        logging.info('Creating data connectors...')
        data_connector_ids.append(
            create_data_connector(supergraph_name, subgraph_name, connector_name, subgraph_data_connector, scalar_ids,
                                  collection_ids, entities))
        logging.info('Finished creating data connectors. Total created: %d', len(data_connector_ids))
        update_entities(entities, include, exclude)
    entities.append({
        "typeName": "subgraph",
        "attributes": {
            "qualifiedName": f"{supergraph_name}.{subgraph_name}",
            "name": subgraph_name,
            "queries": [{"typeName": "query", "uniqueAttributes": {"qualifiedName": item}} for item in
                        query_ids],
            "mutations": [],
            "data_connectors": [{"typeName": "data_connector", "uniqueAttributes": {"qualifiedName": item}} for item in
                                data_connector_ids],
            "subgraph_schema": [{"typeName": "object_type", "uniqueAttributes": {"qualifiedName": item}} for item in
                                object_type_ids]
        }
    })
    update_entities(entities, include, exclude)
    logging.info('Finished generating subgraph...')
    return entities



import json
import os

from ddnatlas.claude import process_json_with_claude
from ddnatlas.get_entities import get_entities

descriptions_prompt = """This is an array of Apache Atlas entities: {{entities}}. For each entity generate a 
description based on the entity's metadata and it's context.

Add an entry to a dictionary where the key is the "qualifiedName" of the entity and the value is the derived 
description. In this format:

{ "<entity qualified name">: "<description>" }
"""


def add_descriptions(entities, include, exclude):
    api_key = os.environ["ANTHROPIC_API_KEY"]

    if include and 'descriptions' not in include:
        return
    if exclude and 'descriptions' in exclude:
        return
    if not api_key:
        return

    existing_entities = get_entities()
    updated_entities = []
    missing_descriptions = [item for item in existing_entities if
                            item['attributes'].get("description") is None or item['attributes'].get(
                                'description') == ""]
    new_descriptions = process_json_with_claude(
        descriptions_prompt.replace('{{entities}}', json.dumps(missing_descriptions)),
        api_key
    )
    for key, description in new_descriptions.items():
        for entity in entities:
            if entity["attributes"]["qualifiedName"] == key:
                entity["attributes"]["description"] = description
                updated_entities.append(entity['attributes'].get('qualifiedName'))
    return updated_entities



def create_scalars(entities, supergraph_name, subgraph_name, connector_name, subgraph_data_connector):
    scalar_ids = []
    for scalar_name, scalar in subgraph_data_connector['schema']['scalar_types'].items():
        qualified_name = f"{supergraph_name}.{subgraph_name}.{connector_name}.{scalar_name}"
        scalar_ids.append(qualified_name)
        entities.append({
            "typeName": "scalar",
            "attributes": {
                "name": scalar_name,
                "representation": scalar['representation']['type'],
                "aggregate_functions": list(scalar['aggregate_functions'].keys()),
                "comparison_operators": list(scalar['comparison_operators'].keys()),
                "qualifiedName": qualified_name
            }
        })
    return scalar_ids



import json
import logging
import os
import click
from pyfiglet import Figlet
from dotenv import load_dotenv

from ddnatlas.get_entities import get_entities
from ddnatlas.supergraph_atlas_types import generate_supergraph_types
from ddnatlas.update import update_supergraph_metadata

# setting up logging
logging.basicConfig(level=logging.INFO,
                    format="%(asctime)s [%(levelname)s] %(message)s",
                    handlers=[
                        logging.FileHandler("../debug.log"),
                        logging.StreamHandler()
                    ])
atlas_logger = logging.getLogger('apache_atlas')
atlas_logger.setLevel(logging.DEBUG)

# Load .env file if it exists
load_dotenv()


def generate_banner(text="DDN ATLAS", font="slant"):
    f = Figlet(font=font)
    return f.renderText(text)


class BannerGroup(click.Group):
    def __call__(self, *args, **kwargs):
        click.echo(generate_banner())
        return super().__call__(*args, **kwargs)


def update_env(_ctx, param, value):
    if value is not None:
        os.environ[param.name.upper()] = value
    return value


def display_configuration():
    click.echo("Current configuration:")
    for key in ['ATLAS_URL', 'SUPERGRAPH', 'ATLAS_USERNAME', 'ATLAS_PASSWORD',
                'ANTHROPIC_API_KEY', 'ANTHROPIC_URI', 'ANTHROPIC_VERSION']:
        value = os.environ.get(key, "Not defined")
        if key in ['ATLAS_PASSWORD', 'ANTHROPIC_API_KEY']:
            value = '********' if value != "Not defined" else value
        click.echo(f"{key}: {value}")


@click.group(cls=BannerGroup)
@click.version_option(version="1.0.0")
@click.option('--atlas-url', required=True, envvar='ATLAS_URL', callback=update_env,
              help='Atlas URL (required if ATLAS_URL env var is not set)')
@click.option('--supergraph', required=True, envvar='SUPERGRAPH', callback=update_env,
              help='Supergraph identifier (required if SUPERGRAPH env var is not set)')
@click.option('--atlas-username', default='admin', envvar='ATLAS_USERNAME', callback=update_env,
              show_default='admin or ATLAS_USERNAME env var', help='Atlas username')
@click.option('--atlas-password', default='admin', envvar='ATLAS_PASSWORD', callback=update_env,
              show_default='admin or ATLAS_PASSWORD env var', help='Atlas password')
@click.option('--anthropic-api-key', envvar='ANTHROPIC_API_KEY', callback=update_env,
              help='Anthropic API Key')
@click.option('--anthropic-uri', default='https://api.anthropic.com', envvar='ANTHROPIC_URI', callback=update_env,
              show_default='https://api.anthropic.com or ANTHROPIC_URI env var', help='Anthropic API URI')
@click.option('--anthropic-version', default='2023-06-01', envvar='ANTHROPIC_VERSION', callback=update_env,
              show_default='2023-06-01 or ANTHROPIC_VERSION env var', help='Anthropic API Version')
def cli(atlas_url, supergraph, atlas_username, atlas_password, anthropic_api_key, anthropic_uri, anthropic_version):
    """
    DDN Atlas CLI - A command-line tool for managing supergraph metadata within Apache Atlas.

    This tool provides commands to initialize configuration, update metadata,
    and download supergraph metadata from Apache Atlas.
    """
    pass


@cli.command()
def init():
    """
    Initialize the configuration for DDN Atlas.

    This command updates Apache Atlas with the supergraph metadata types.
    """
    display_configuration()
    click.echo("Adding supergraph metadata types to Atlas...")
    generate_supergraph_types()


@click.option(
    '--include',
    '-i',
    multiple=True,
    default=None,
    help=('Include specific components in the update. '
          'Options: subgraph, supergraph, entity, relationship, business_metadata, '
          'glossary, data_connector, model, object_type, scalar, descriptions')
)
@click.option(
    '--exclude',
    '-e',
    multiple=True,
    default=None,
    help=('Exclude specific components from the update. '
          'Options: subgraph, supergraph, entity, relationship, business_metadata, '
          'glossary, data_connector, model, object_type, scalar, descriptions')
)
@cli.command()
def update(include, exclude):
    """
    Update the supergraph metadata in Apache Atlas.

    This command allows you to update the supergraph metadata stored in Apache Atlas.
    It uses the configuration settings to connect to Atlas and update the specified
    supergraph's metadata.
    """
    display_configuration()
    click.echo("Updating the supergraph metadata to atlas...")
    include = list(include) if len(include) > 0 else None
    exclude = list(exclude) if len(exclude) > 0 else None
    update_supergraph_metadata(include, exclude)


@cli.command()
@click.option('--output', '-o', help='Output file path')
def dump(output):
    """
    Download the Apache Atlas supergraph metadata.

    This command retrieves the supergraph metadata from Apache Atlas and saves it
    to a file. If no output file is specified, it will display the metadata in the console.
    """
    display_configuration()
    click.echo("Downloading supergraph metadata from Apache Atlas...")
    if output:
        with open(output, 'w') as f:
            json.dump(get_entities(), f, indent=2)
        click.echo(f"Supergraph metadata has been written to {output}")
    else:
        print(json.dumps(get_entities(), indent=2))





import os

from apache_atlas.client.base_client import AtlasClient
from ddnatlas.camel_to_title import camel_to_title


def add_business_metadata(entities, include, exclude):

    url = os.getenv('ATLAS_URL')
    username = os.getenv('ATLAS_USERNAME')
    password = os.getenv('ATLAS_PASSWORD')

    if include and 'business_metadata' not in include:
        return
    if exclude and 'business_metadata' in exclude:
        return

    client = AtlasClient(host=url, auth=(username, password))
    for entity in entities:
        if entity.get('typeName') and entity['attributes'].get('qualifiedName') and entity['attributes'].get(
                'name') and entity.get('typeName') in ["field", "column", "query", "collection", "subgraph",
                                                       "supergraph", "object_type"]:
            qualified_name = entity['attributes']['qualifiedName']
            technical_name = entity['attributes']['name']
            business_name = camel_to_title(technical_name.replace('_', ' '))
            result = client.discovery.attribute_search(type_name='Asset', attr_name='qualifiedName',
                                                       attr_value_prefix=qualified_name, limit=1, offset=0)
            guid = result['entities'][0]['guid']
            business_metadata = {
                "data_analysis": {
                    "businessNames": [business_name]
                }
            }
            try:
                client.entity.add_or_update_business_attributes(entity_guid=guid, is_overwrite=False,
                                                                business_attributes=business_metadata)
            except Exception:
                pass



import requests
import logging
from requests.auth import HTTPBasicAuth


def execute_request(http_method, url, headers={}, params=None, json_data=None, username=None, password=None):
    auth = None
    if username and password:
        auth = HTTPBasicAuth(username, password)

    try:
        if http_method == "get":
            logging.info('Sending GET request to %s', url)
            response = requests.get(url, params=params, headers=headers, auth=auth)
        elif http_method == "post":
            logging.info('Sending POST request to %s', url)
            response = requests.post(url, params=params, headers=headers, json=json_data, auth=auth)
        elif http_method == "delete":
            logging.info('Sending DELETE request to %s', url)
            response = requests.delete(url, params=params, headers=headers, auth=auth)
        elif http_method == "put":
            logging.info('Sending PUT request to %s', url)
            response = requests.put(url, params=params, headers=headers, json=json_data, auth=auth)
        else:
            return None

        logging.info('Response status code: %d', response.status_code)
        return response

    except Exception:
        logging.error('Exception occurred', exc_info=True)



def create_columns(entities, supergraph_name, subgraph_name, connector_name, collection_name, object_schema):
    column_ids = []
    for column_name, column in object_schema['fields'].items():
        nullable = column['type'].get('type') == 'nullable'
        if nullable:
            type_name = column['type']['underlying_type']['name']
        else:
            type_name = column['type']['name']
        scalar_name = f"{supergraph_name}.{subgraph_name}.{connector_name}.{type_name}"
        qualified_name = f"{supergraph_name}.{subgraph_name}.{collection_name}.{column_name}"
        column_ids.append(qualified_name)
        entities.append({
            "typeName": "column",
            "attributes": {
                "name": column_name,
                "qualifiedName": qualified_name,
                "nullable": nullable,
                "dataType": {"typeName": "scalar", "uniqueAttributes": {"qualifiedName": scalar_name}},
            }
        })
    return column_ids



import logging
import os

from ddnatlas.execute_request import execute_request


def get_object_type(object_types, name):
    return next(item for item in object_types if item['definition']['name'] == name)


def get_entity(entities, _name, object_type):
    return next(entity for entity in entities if
                entity['typeName'] == 'object_type' and entity['attributes']['name'] ==
                object_type['definition']['graphql']['typeName'])


def get_key_ids(entity, keys):
    key_ids = []
    for key in keys:
        key_ids += [field['uniqueAttributes']['qualifiedName'] for field in entity['attributes']['fields'] if
                    field['uniqueAttributes']['qualifiedName'].endswith('.' + key)]
    return key_ids


def create_atlas_relationship(relationship, source_entity, target_entity, source_key_ids, target_key_ids):
    return {
        "typeName": "supergraph_array_relationship" if relationship['definition']['target']['model'][
                                                           'relationshipType'] != 'Object' else "supergraph_object_relationship",
        "attributes": {
            "field_name": relationship['definition']['name'],
            "source_keys": [{
                "typeName": "column",
                "uniqueAttributes": {
                    "qualifiedName": item
                }
            } for item in source_key_ids],
            "target_keys": [{
                "typeName": "column",
                "uniqueAttributes": {
                    "qualifiedName": item
                }
            } for item in target_key_ids]
        },
        "end1": {
            "typeName": "object_type",
            "uniqueAttributes": {
                "qualifiedName": source_entity['attributes']['qualifiedName']
            }
        },
        "end2": {
            "typeName": "object_type",
            "uniqueAttributes": {
                "qualifiedName": target_entity['attributes']['qualifiedName']
            }
        },
        "status": "ACTIVE"
    }


def create_relationships(relationships, object_types, entities):
    atlas_url = os.getenv('ATLAS_URL')
    username = os.getenv('ATLAS_USERNAME')
    password = os.getenv('ATLAS_PASSWORD')

    logging.info("Creating relationships..")

    for relationship in relationships:
        source_object_type = get_object_type(object_types, relationship['definition']['sourceType'])
        target_object_type = get_object_type(object_types,
                                             relationship['definition']['target']['model']['name'])

        source_entity = get_entity(entities, source_object_type['definition']['graphql']['typeName'],
                                   source_object_type)
        target_entity = get_entity(entities, target_object_type['definition']['graphql']['typeName'],
                                   target_object_type)

        source_keys = ['.'.join(item['fieldName'] for item in mapping['source']['fieldPath']) for mapping in
                       relationship['definition']['mapping']]
        target_keys = ['.'.join(item['fieldName'] for item in mapping['target']['modelField']) for mapping in
                       relationship['definition']['mapping']]

        source_key_ids = get_key_ids(source_entity, source_keys)
        target_key_ids = get_key_ids(source_entity, target_keys)

        atlas_relationship = create_atlas_relationship(relationship, source_entity, target_entity, source_key_ids,
                                                       target_key_ids)
        execute_relationship_request(atlas_url, username, password, atlas_relationship)
    supergraph = [item for item in entities if item.get('typeName') == 'supergraph'][0]
    for entity in entities:
        if entity['typeName'] == 'field':
            rel = {
                "typeName": "data_source_to_api",
                "attributes": {
                    "transform": f"{supergraph['attributes'].get('name', 'enterprise')}.straight_through"
                },
                "end1": {
                    "typeName": "column",
                    "uniqueAttributes": {
                        "qualifiedName": entity['attributes']['column']['uniqueAttributes']['qualifiedName']
                    }
                },
                "end2": {
                    "typeName": "field",
                    "uniqueAttributes": {
                        "qualifiedName": entity['attributes']['qualifiedName']
                    }
                },
                "status": "ACTIVE"
            }
            execute_relationship_request(atlas_url, username, password, rel)
        if entity['typeName'] == 'collection':
            for column in entity['attributes']['columns']:
                rel = {
                    "typeName": "collection_owns_column",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the column: {column['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "collection",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "column",
                        "uniqueAttributes": {
                            "qualifiedName": column['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
        if entity['typeName'] == 'object_type':
            for field in entity['attributes']['fields']:
                rel = {
                    "typeName": "object_type_owns_field",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the field: {field['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "object_type",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "field",
                        "uniqueAttributes": {
                            "qualifiedName": field['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
        if entity['typeName'] == 'data_connector':
            for collection in entity['attributes']['collections']:
                rel = {
                    "typeName": "data_connector_has_collection",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the collection: {collection['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "data_connector",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "collection",
                        "uniqueAttributes": {
                            "qualifiedName": collection['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
        if entity['typeName'] == 'supergraph':
            for subgraph in entity['attributes']['domains']:
                rel = {
                    "typeName": "supergraph_has_subgraph",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the collection: {subgraph['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "supergraph",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "subgraph",
                        "uniqueAttributes": {
                            "qualifiedName": subgraph['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
        if entity['typeName'] == 'subgraph':
            for data_connector in entity['attributes']['data_connectors']:
                rel = {
                    "typeName": "subgraph_has_data_connector",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the collection: {data_connector['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "subgraph",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "data_connector",
                        "uniqueAttributes": {
                            "qualifiedName": data_connector['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
            for query in entity['attributes']['queries']:
                rel = {
                    "typeName": "subgraph_has_query",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the collection: {query['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "subgraph",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "query",
                        "uniqueAttributes": {
                            "qualifiedName": query['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
            for mutation in entity['attributes']['mutations']:
                rel = {
                    "typeName": "subgraph_has_mutation",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the collection: {mutation['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "subgraph",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "mutation",
                        "uniqueAttributes": {
                            "qualifiedName": mutation['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
            for object_type in entity['attributes']['subgraph_schema']:
                rel = {
                    "typeName": "subgraph_has_object_type",
                    "attributes": {
                        "description": f"{entity['attributes']['qualifiedName']} has the object_type: {object_type['uniqueAttributes']['qualifiedName']}"
                    },
                    "end1": {
                        "typeName": "subgraph",
                        "uniqueAttributes": {
                            "qualifiedName": entity['attributes']['qualifiedName']
                        }
                    },
                    "end2": {
                        "typeName": "object_type",
                        "uniqueAttributes": {
                            "qualifiedName": object_type['uniqueAttributes']['qualifiedName']
                        }
                    },
                    "status": "ACTIVE"
                }
                execute_relationship_request(atlas_url, username, password, rel)
        if entity['typeName'] == 'query':
            object_type = entity['attributes']['record']
            rel = {
                "typeName": "query_has_object_type",
                "attributes": {
                    "description": f"{entity['attributes']['qualifiedName']} has the collection: {object_type['uniqueAttributes']['qualifiedName']}"
                },
                "end1": {
                    "typeName": "query",
                    "uniqueAttributes": {
                        "qualifiedName": entity['attributes']['qualifiedName']
                    }
                },
                "end2": {
                    "typeName": "object_type",
                    "uniqueAttributes": {
                        "qualifiedName": object_type['uniqueAttributes']['qualifiedName']
                    }
                },
                "status": "ACTIVE"
            }
            execute_relationship_request(atlas_url, username, password, rel)


def execute_relationship_request(atlas_url, username, password, atlas_relationship):
    logging.info("Executing relationship request..")
    response = execute_request('post', f"{atlas_url}/api/atlas/v2/relationship", json_data=atlas_relationship,
                               username=username, password=password)
    if response.status_code not in (200, 409):
        raise Exception(response.text)



apache_primitives = {
    "int32": "int",
    "float32": "float",
    "string": "string",
    "timestamp": "date",
    "BIGINT": "biginteger",
    "FLOAT": "float",
    "INTEGER": "long",
    "VARCHAR": "string",
    "DOUBLE": "double",
    "DECIMAL": "bigdecimal",
    "TIMESTAMP": "date",
    "TIMESTAMPTZ": "date",
    "DATE": "date",
    "DATETIME": "date",
    "CHAR": "byte",
    "BOOLEAN": "boolean",
    "VARBINARY": "string",
    "JSON": "string",
    "LIST": "string",
    "MAP": "string"
}



def get_object_type_and_name(object_types, connector_name, collection_name):
    object_type = [item for item in object_types if any(
        obj.get('dataConnectorName') == connector_name and obj.get('dataConnectorObjectType') == collection_name
        for obj in item['definition']['dataConnectorTypeMapping'])][0]
    object_name = object_type['definition']['graphql']['typeName']

    return object_type, object_name


def create_object_type(supergraph_name, subgraph_name, object_name, field_ids, entities):
    object_qualified_name = "{}.{}.{}".format(supergraph_name, subgraph_name, object_name)
    object_type_info = {
        "typeName": "object_type",
        "attributes": {
            "name": object_name,
            "qualifiedName": object_qualified_name,
            "fields": [{
                "typeName": "field",
                "uniqueAttributes": {
                    "qualifiedName": item
                }
            } for item in field_ids]
        }
    }
    entities.append(object_type_info)
    return object_qualified_name



import json
import os
import anthropic


def process_json_with_claude(prompt, api_key=None):
    client = anthropic.Client(api_key=api_key)
    message = client.messages.create(
        model="claude-3-5-sonnet-20240620",
        max_tokens=8192,
        temperature=0,
        system="""You are a world-class data scientist at a Fortune 500 company specializing in data governance and 
        metadata management. Always respond with valid JSON objects. Each response should be a single JSON object 
        containing the requested information. Do not include any text outside of the JSON object in your response.""",
        messages=[
            {
                "role": "user",
                "content": [
                    {
                        "type": "text",
                        "text": prompt
                    }
                ]
            }
        ]
    )
    """
    Process a JSON structure using the Claude 3.5 Sonnet API and return a JSON structure.
    :param input_json: Dict or JSON-serializable object to be processed
    :param prompt: String containing the prompt for Claude
    :param api_key: Optional API key. If not provided, it will be read from the ANTHROPIC_API_KEY environment variable
    :return: Processed JSON structure
    """
    if api_key is None:
        api_key = os.environ.get('ANTHROPIC_API_KEY')
        if api_key is None:
            raise ValueError("API key not provided and ANTHROPIC_API_KEY environment variable not set")

    try:
        full_response = ''.join(block.text for block in message.content)
        return json.loads(full_response)
    except anthropic.AnthropicError as e:
        raise Exception(f"API request failed with status code {e}")
    except json.JSONDecodeError:
        raise ValueError("The API response could not be parsed as JSON")



import logging
import os
from typing import Optional, List

from apache_atlas.client.base_client import AtlasClient
from apache_atlas.model.instance import AtlasEntity, AtlasEntitiesWithExtInfo, AtlasEntityWithExtInfo
from apache_atlas.utils import type_coerce

already_processed = set()


def include_element(element: str, include: Optional[List[str]], exclude: Optional[List[str]]):
    if not include and not exclude:
        return True
    elif exclude and element in exclude:
        return False
    elif include and element not in include:
        return False
    return True


def remove_nested_key(dictionary, unwanted_key):
    if isinstance(dictionary, dict):
        for key in list(dictionary.keys()):  # We use list() to avoid 'dictionary size changed during iteration' error
            if key == unwanted_key:
                del dictionary[key]
            elif isinstance(dictionary[key], dict):
                remove_nested_key(dictionary[key], unwanted_key)
            elif isinstance(dictionary[key], list):
                for item in dictionary[key]:
                    remove_nested_key(item, unwanted_key)
    return dictionary


def compare_dictionaries(dict1, dict2):
    if set(dict1.keys()) != set(dict2.keys()):
        return False

    for key, val1 in dict1.items():
        val2 = dict2[key]

        if type(val1) is not type(val2):
            return False

        if isinstance(val1, dict):
            if not compare_dictionaries(val1, val2):
                return False
        elif isinstance(val1, list):
            if sorted([str(el) for el in val1]) != sorted([str(el) for el in val2]):
                return False
        elif val1 != val2:
            return False

    return True


def deep_merge(dict1, dict2):
    merged = dict1.copy()
    for key, value in dict2.items():
        if isinstance(value, dict) and key in merged and isinstance(merged[key], dict):
            merged[key] = deep_merge(merged[key], value)
        else:
            merged[key] = value
    return merged


def update_entities(entities: Optional[List[dict]],
                    include: Optional[List[str]] = None, exclude: Optional[List[str]] = None,
                    force_updates: Optional[List[str]] = None):
    atlas_url = os.getenv('ATLAS_URL')
    username = os.getenv('ATLAS_USERNAME')
    password = os.getenv('ATLAS_PASSWORD')

    if force_updates:
        for item in force_updates:
            already_processed.remove(item)
    client = AtlasClient(atlas_url, auth=(username, password))
    for entity in entities:
        if not entity['attributes'].get('qualifiedName') in already_processed and include_element(
                entity.get('typeName'), include, exclude):
            already_processed.add(entity['attributes'].get('qualifiedName'))
            try:
                entity_type = entity['typeName']
                attribute = {'qualifiedName': entity['attributes']['qualifiedName']}.items()
                retrieved_entity = None
                try:
                    retrieved_entity = client.entity.get_entity_by_attribute(
                        type_name=entity_type,
                        uniq_attributes=attribute
                    )
                except Exception:
                    pass

                if retrieved_entity is not None:
                    entity_attributes = entity["attributes"]
                    retrieved_entity = retrieved_entity['entity']
                    retrieved_attributes = retrieved_entity.attributes
                    retrieved_selected_attributes = {k: v for k, v in retrieved_attributes.items() if
                                                     k in entity_attributes}
                    retrieved_selected_attributes = remove_nested_key(retrieved_selected_attributes, 'guid')

                    # if they are not same
                    if not compare_dictionaries(entity_attributes, retrieved_selected_attributes):
                        atlas_entity = type_coerce({"entity": entity}, AtlasEntityWithExtInfo)
                        response = client.entity.update_entity(atlas_entity)
                        logging.info({'action': 'update', 'status': response.status, 'entity': entity})
                    else:
                        logging.info({'action': 'skip', 'status': 'new entity attributes match with existing entity',
                                      'entity': entity})
                else:
                    atlas_entity = AtlasEntity({"typeName": entity.get('typeName')})
                    atlas_entity.attributes = entity.get('attributes')
                    entities_info = AtlasEntitiesWithExtInfo()
                    entities_info.entities = [atlas_entity]
                    response = client.entity.create_entities(entities_info)
                    logging.info({'action': 'create', 'status': response.status, 'entity': entity})
            except Exception as e:
                logging.error({'action': 'error', 'status': str(e), 'entity': entity})



glossary_prompt = """
This is an array of Apache Atlas entities: {{entities}}. 

Review all entities and derive a single glossary entry that encompasses the principal topics referenced in the 
entities. Add this into the response as the value of the "glossary" key.

Review all entities and derive a list of relevant terms and relevant categories.

Associate the terms with a category by the category guid and with a glossary by the glossary guid.
Associate the categories with a glossary by the glossary guid.

Add the terms into the response as the value of the "terms" key.

Add the categories into the response as the value of the "categories" key.

Associate the terms with one or more collections, object_types, queries, columns, or fields using an 
AtlasGlossarySemanticAssignment.

Add the term-entity associations into the response as the value of the "relationships" key.

Review field and column entities and assign any relevant data governance classifications, then add those to the 
"classification" entry. The classification entry should be a dictionary of entity qualified names whose value is an 
array of "PII","Confidential","Internal","Public","Restricted".

Use these definitions when considering classification:

Public Data: Information that is freely available to the public and poses no risk if disclosed. Examples include 
press releases and publicly available reports.

Internal Data: Data intended for internal use within the organization. While not highly sensitive, it should still be 
protected to avoid misuse. Examples include internal memos and internal project documents.

Confidential Data: Sensitive information that, if disclosed, could harm the organization or individuals. This data 
requires strict access controls. Examples include employee records, financial data, and proprietary information.

Restricted Data: Highly sensitive data that requires the highest level of protection due to its critical nature. 
Unauthorized access could lead to severe consequences. Examples include trade secrets, classified government 
information, and personal health information (PHI).

Personally Identifiable Information (PII): Data that can be used to identify an individual, either on its own or when 
combined with other information. This includes names, addresses, social security numbers, email addresses, 
and phone numbers. PII requires stringent protection measures to prevent identity theft and ensure privacy.

Enter the classifications into the response as the value of the "classifications": key.

The complete answer must be a JSON dictionary following this format:

{ "glossary": { "guid": "<temporary guid as a unique negative number>", "qualifiedName": "<<the name of the glossary 
in title case without spaces>@{{supergraph_name}}>", "name": "<the name of the glossary>", "shortDescription": "<a 
short description of the glossary>", "longDescription": "<a longer description of the glossary>", "language": 
"English", "usage": "<a short description of when to use this glossary>" }, "categories": [ { "guid": "<temporary 
guid as a unique negative number>", "qualifiedName": "<<the name of the term in category in title case without 
spaces>@.<glossary name in lower case>.{{supergraph_name}}>", "name": "<the category name>", "shortDescription": "<a 
short description of the category>", "longDescription": "<a long description of the category>", "anchor": { 
"glossaryGuid": "<the guid of the glossary>" } } ], "terms": [ { "guid": "<temporary guid as a unique negative 
number>", "qualifiedName": "<<the name of the term in title case without spaces>@.<glossary name in lower case>.{{
supergraph_name}}>", "name": "<the term>", "shortDescription": "<a short description of the term>", 
"longDescription": "<a long description of the term>", "examples": [ "<an array of examples of using the term>" ], 
"abbreviation": "<an abbreviation of the term>", "usage": "<describe the scenarios where the term might be used>", 
"status": "ACTIVE", "anchor": { "glossaryGuid": "<the guid of the glossary>" }, "categories": [ { "categoryGuid": 
"<the category guid associated with the term>" } ] } ], "relationships": [ { "typeName": 
"AtlasGlossarySemanticAssignment", "end1": { "typeName": "AtlasGlossaryTerm", "guid": "<glossary_term_guid>" }, 
"end2": { "typeName": "<entity_type_name>", "uniqueAttributes": { "qualifiedName": "<entity_qualified_name>" } } } ], 
"classifications": { "<entity qualified name>": [ "<classification category>" ] } }"""



import os
from typing import List, Dict, Any
import logging
import yaml


def parse_files(directory: str) -> List[Dict[str, Any]]:
    parsed_data = []
    logging.info(f"Searching for files in: {os.path.abspath(directory)}")
    for root, dirs, files in os.walk(directory):
        logging.info(f"Current directory: {root}")
        logging.info(f"Subdirectories: {dirs}")
        logging.info(f"Files: {files}")
        for file in files:
            if file.endswith(('.hml', '.yml', '.yaml')):
                file_path = os.path.join(root, file)
                logging.info(f"Processing file: {file_path}")
                try:
                    with open(file_path, 'r') as f:
                        content = f.read()
                        if content.startswith('---'):
                            # Multi-document YAML file
                            docs = yaml.safe_load_all(content)
                            parsed_data.extend(list(docs))
                        else:
                            # Single YAML document
                            parsed_data.append(yaml.safe_load(content))
                    logging.info(f"Successfully parsed: {file_path}")
                except Exception as e:
                    logging.error(f"Error processing {file_path}: {e}")

    logging.info(f"Total parsed documents: {len(parsed_data)}")
    return parsed_data



def create_queries(supergraph_name, subgraph_name, object_qualified_name, data_connector_model, entities):
    query_ids = []
    select = data_connector_model['definition']['graphql'].get('selectMany').get('queryRootField')
    qualified_name = f'{supergraph_name}.{subgraph_name}.{select}'
    entities.append({
        "typeName": "query",
        "attributes": {
            "name": select,
            "qualifiedName": qualified_name,
            "is_array": True,
            "unique_identifiers": [],
            "record": {
                "typeName": "object_type",
                "uniqueAttributes": {
                    "qualifiedName": object_qualified_name
                }
            }
        }
    })
    query_ids.append(qualified_name)
    select_uniques = [item for item in data_connector_model['definition']['graphql'].get('selectUniques')]
    for selectUnique in select_uniques:
        qualified_name = f'{supergraph_name}.{subgraph_name}.{selectUnique["queryRootField"]}'
        entities.append({
            "typeName": "query",
            "attributes": {
                "name": selectUnique["queryRootField"],
                "qualifiedName": qualified_name,
                "is_array": False,
                "unique_identifiers": selectUnique["uniqueIdentifier"],
                "record": {
                    "typeName": "object_type",
                    "uniqueAttributes": {
                        "qualifiedName": object_qualified_name
                    }
                }
            }
        })
        query_ids.append(qualified_name)
    return query_ids
